(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/StaticScope.sml,v
   Comment: Machinery to compute static scope for transitions. If needed
	    transition is multiplied to eliminate dynamic scope.
   Release: scope-0-10

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The module assumes that there are only event transitions coming in. GROUP
   TRIGGERED TRANSITION SHOULD BE FIRST MULTIPLIED WITH Normalise.degroup.

   1. scope of force state action to state which is already active is not
   correct. Behaves as self loop. Actually this is inconsistent in VS. The
   explicit target behaves as self loop while force state targets have
   internal scope. Yuck.

   StaticScope.sml,v 1.15 2002/10/30 21:21:31 wasowski Exp
*)

structure StaticScope = struct

  open Haskellish Absyn AbsynLogic Err Say
  infixr 0 $
  infix 9 sub

  (* type of static transitions with scopes embedded *)

  type stran = { trigger :ebind,
		 pos	 :andstate list,
		 neg	 :andstate list,
		 guard	 :bexp list,
		 action  :action,
		 targets :(andstate list * orstate) list } (* (targets,scope)list *)

  (* compute the minLCA - lowest state being ancestor to at least one
     positive condition and all the targets of transition *)
  
  fun mLCA (h :hierarchy) (t :tran) :andstate = 
	let val ts = map nameand (#targets t)
	    val ss = map (fn p => irrLCA h (nameand p::ts)) (#pos t)
	    fun f (s,mlca) = if ancestorh h s mlca then mlca else s
	    val mlca = foldl f (hd ss) (tl ss) 
		       handle Empty => impossible "tran should have 1+ poscond"
	    val mlca1 = getstate h mlca
	    val mlca2 = if isor mlca1 then getstate h $ parent mlca1 
			else mlca1
	in valOf $ getand mlca2 end

  (* build bdd of part of the statechart below minLCA, restrict it for
     positive and negative conditions and existentially quantify on
     nonrelevant variables. The bdd returned should have support
     only in orstates between targets and mLCA *)
  
  fun solve (h :hierarchy) (t :tran, mlca :andstate) 
		:(stateid list * stateid list) list = 
	let val tgts = map nameand (#targets t)
	    val pos  = map nameand (#pos t)
	    val neg  = map nameand (#neg t)
	    val _    = sayln Verbose ("mLCA: " ^ Symbol.name (nameand mlca))
	    val r    = AbsynLogic.build (InclAND mlca)
	    val _    = sayln Verbose (AbsynLogic.sprMapping())
	    val _    = IfVerbose(fn()=> bdd.fnprintdot "_hierarchy_r.dot" r)
	    val r1   = AbsynLogic.force r (pos,neg)
	    val _    = IfVerbose(fn()=> bdd.fnprintdot "_hierarchy_r1.dot" r1)
	    val ancs = List.filter (not o ancestorhst h (nameor $ #scope t)) 
				$ List.filter (ancestorhst h $ nameand mlca)
				$ List.filter (isandh h) 
				$ List.concat 
				$ map (ancesth h) tgts
	    val _    = sayln Verbose ("Number of targets in restriction step: " ^
				    Int.toString (length tgts))
	    val _    = sayln Verbose ("Number of states we look at: " ^
				    Int.toString (length ancs))
	    val r2   = AbsynLogic.restrict_to r1 ancs
	    val _    = IfVerbose(fn()=> bdd.fnprintdot "_hierarchy_r2.dot" r2)
	    val soluts = AbsynLogic.allsat r2 ancs
	in soluts end


  (* Split targets of transition into ancestrally related groups with
     single scope *)
     
  fun groupByScope (ts :(andstate * orstate) list) 
				:(andstate list * orstate) list = 
	case ts of [] => []
		 | (a1,o1)::ta =>
			let val (grp,rest) = 
				List.partition (fn (a2,o2) => o1 = o2) ta
			in (a1::map #1 grp, o1) :: groupByScope rest end


   (* make new scope for single target. pos is assumed to be not simplified. 
      Should be not less than what is returned from bdd pacakge.  It is then 
      suffcient to find the nearest ancestor on list of positive conditions. 
      returns target and its scope.
      
      es - explicit scope of transition. Used for uncovering algorithm *)

  fun staticscope (h :hierarchy) (pos :andstate list) 
			(tgt :andstate) :andstate * orstate =
	let val npos = map nameand pos
	    val nt   = nameand tgt
	    val root = name (getroot h)
	    fun lowest (p,e) = 
		if ancestorh h p nt then if ancestorh h e p then p else e else e
	    val s    = foldl lowest root npos
	    val ors  = map nameor $ childrenand (valOf $ getand $ getstate h s)
	    val sc   = valOf $ List.find (fn s => ancestorh h s nt) ors
	    val _    = sayln Verbose ("Static scope found: "^ Symbol.name sc)
	in (tgt,valOf $ getor $ getstate h sc) end 
	
  (* Append newly computed pos and negconds and voila.
     you have a static transition. cleanup propagation should be done *)
     
  fun mk_stran (h :hierarchy) (t :tran) [(pos,neg)] :stran list =
  
	(* do not add anything to conditions if there is only a single
	   solution. A simple workaround. General should do in all cases:
	   never add a condition if it is the same in all new cases. FIXME *)

	let val (pos1,neg1) = Simplifier.simplify h (#pos t, #neg t)
	    val ts          = map (staticscope h pos) (#targets t)
	in [{ trigger = getebind (#trigger t), 
	     pos = pos1, neg = neg1,
	     guard   = #guard t, action = #action t,
	     targets = groupByScope ts }]
	end

    | mk_stran (h :hierarchy) (t :tran) soluts =
	let fun f (pos,neg) =
		let val (pos1,neg1) = (pos @ #pos t, neg @ #neg t)
		    val (pos2,neg2) = Simplifier.simplify h (pos1,neg1)
		    val ts = map (staticscope h pos) (#targets t)
		in
		   { trigger = getebind (#trigger t),
		     pos     = pos2, neg = neg2, 
		     guard   = #guard t, action  = #action t,
		     targets = groupByScope ts } 
		end
	in map f soluts end


  (* This function is to be called externally. If transition is not to be
     fired ever it is eliminated. Otherwise static scope is computed
     (multiplying transitions if needed) and list of equivalent static
     transitions is returned.
     
     Perhaps the next function is even more usefull for calling externally *)
  
  fun static_scopes (h :hierarchy) (t :tran) :stran list = 
	let val t1     = Simplifier.simplify_tran h t
	    val mlca   = mLCA h t1
	    val soluts = solve h (t1,mlca)
	    val _      = sayln Verbose "Solutions BEGIN"
	    val _      = map (sayln Verbose o sprSolution) soluts
	    val _      = sayln Verbose "Solutions END"
	    val aa     = hd soluts
	    val conds1 = map (fn (p,n) => (map (valOf o getand o getstate h) p,
					   map (valOf o getand o getstate h) n)) 
					   soluts
	in mk_stran h t conds1 end

  (* compute static scopes for all transitions in one shot *)

  fun static_scopes_all (h :hierarchy) (trans :tran list) :stran list =
	List.concat $ map (static_scopes h) trans

  (* diagnostic/debugging/testing support *)

  local 
  
	open AbsynPrint Typeset
  in

	fun sprTgts (ta :andstate list, sc : orstate) :string =
		String.concat ("[" :: sprStateor sc :: "] " 
			:: join_blanks(map sprStateand ta) [" "])

	fun sprSTran {trigger,pos,neg,guard,action,targets} :string =
		String.concat ( sprEBind trigger :: " : POS"
			:: c_params(map sprStateand pos) (" NEG"
				:: c_params(map sprStateand neg) ( " -> "
					:: map sprTgts targets )))
	
	fun prSTranln stran = println (sprSTran stran)
  end
 

end
